local floor = math.floor
local deg = math.deg
local GetUnitSpeed = GetUnitSpeed
local GetUnitPitch = GetUnitPitch
local IsFlying = IsFlying
local IsSwimming = IsSwimming

local db
local speeder = "player"


-- data for the various unit displays
local unitData = {
	percent = {
		display = "%",
		width = 64,
	},
	yards = {
		display = " yd/s",
		width = 80,
	},
	miles = {
		display = " mph",
		width = 80,
	},
	kilometers = {
		display = " km/h",
		width = 88,
	},
	meters = {
		display = " m/s",
		width = 80,
	},
}


-- create the speed-o-meter frame
local addon = CreateFrame("Frame", "SpeedyGonzalesFrame", UIParent)
addon:SetMovable(true)
addon:SetToplevel(true)
addon:SetHeight(32)
addon:SetBackdrop({
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
	tile = true, tileSize = 16, edgeSize = 16,
	insets = {left = 4, right = 4, top = 4, bottom = 4}
})
addon:SetBackdropColor(0, 0, 0, 0.8)
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_ENTERING_WORLD")
addon:RegisterEvent("UNIT_ENTERED_VEHICLE")
addon:RegisterEvent("UNIT_EXITED_VEHICLE")
addon:SetScript("OnEvent", function(self, event, ...)
	return self[event] and self[event](self, ...)
end)
addon:SetScript("OnMouseDown", addon.StartMoving)
addon:SetScript("OnMouseUp", function(self) self:OnMouseUp() end)
addon:SetScript("OnHide", function(self) self:OnMouseUp() end)
addon:SetScript("OnUpdate", function(self)
	local pitch = ""
	local flying = IsFlying() or IsSwimming()
	if db.pitch then
		if flying then
			pitch = "\n%.1f \194\176"
		else
			pitch = "\n-"
		end
	end
	self.text:SetFormattedText(("%d%s"..pitch), self:GetSpeed(), unitData[db.units].display, deg(GetUnitPitch(speeder)))
end)

-- create font string for the actual speed text
local addonText = addon:CreateFontString()
addonText:SetPoint("CENTER", addon)
addonText:SetFontObject(GameFontHighlight)
addon.text = addonText


local optionsFrame = CreateFrame("Frame")
optionsFrame.name = "SpeedyGonzales"
InterfaceOptions_AddCategory(optionsFrame)


local title = optionsFrame:CreateFontString(nil, nil, "GameFontNormalLarge")
title:SetPoint("TOPLEFT", 16, -16)
title:SetPoint("RIGHT", -16, 0)
title:SetJustifyH("LEFT")
title:SetJustifyV("TOP")
title:SetText(optionsFrame.name)
optionsFrame.title = title


local function onClick(self)
	local checked = self:GetChecked() == 1
	PlaySound(checked and "igMainMenuOptionCheckBoxOn" or "igMainMenuOptionCheckBoxOff")
	db[self.setting] = checked
	if self.func then
		addon[self.func](addon)
	end
end

function optionsFrame:NewOption()
	local button = CreateFrame("CheckButton", nil, self, "OptionsBaseCheckButtonTemplate")
	button:SetPushedTextOffset(0, 0)
	button:SetScript("OnClick", onClick)
	
	local buttonText = button:CreateFontString(nil, nil, "GameFontHighlight")
	buttonText:SetPoint("LEFT", button, "RIGHT", 0, 1)
	button:SetFontString(buttonText)

	return button
end


-- check buttons data
local options = {
	{
		text = "Show",
		setting = "shown",
		func = "SetVisibility",
	},
	{
		text = "Lock",
		setting = "locked",
		func = "SetLock",
	},
	{
		text = "Pitch display",
		setting = "pitch",
		func = "SetPitchDisplay",
	},
}


optionsFrame.options = {}

for i, v in ipairs(options) do
	local option = optionsFrame:NewOption()
	if i == 1 then
		option:SetPoint("TOPLEFT", optionsFrame.title, "BOTTOMLEFT", -2, -16)
	else
		option:SetPoint("TOP", optionsFrame.options[i - 1], "BOTTOM", 0, -8)
	end
	option:SetText(v.text)
	option.setting = v.setting
	option.func = v.func
	optionsFrame.options[i] = option
end


local dropdown = CreateFrame("Frame", "SpeedyGonzalesUnitsMenu", optionsFrame, "UIDropDownMenuTemplate")
dropdown:SetPoint("TOPLEFT", optionsFrame.options[#options], "BOTTOMLEFT", -13, -24)
dropdown.label = dropdown:CreateFontString(nil, "BACKGROUND", "GameFontNormalSmall")
dropdown.label:SetPoint("BOTTOMLEFT", dropdown, "TOPLEFT", 16, 3)
dropdown.label:SetText("Units")


-- slash command opens options frame
SLASH_SPEEDYGONZALES1 = "/speedy"
SlashCmdList["SPEEDYGONZALES"] = function(msg)
	msg = msg:trim()
	if msg:lower() == "config" then
		InterfaceOptionsFrame_OpenToCategory(optionsFrame)
	elseif msg == "" then
		db.shown = not db.shown
		optionsFrame.options[1]:SetChecked(db.shown)
		addon:SetVisibility()
	else
		print("|cffffff00SpeedyGonzales:|r Type '/speedy' to toggle the frame or '/speedy config' to open the configuration.")
	end
end


function addon:ADDON_LOADED(addon)
	if addon == "SpeedyGonzales" then
		-- create options defaults if they do not exist
		db = SpeedyGonzalesData or SpeedyGonzalesDB or {
			units = "percent",
			pitch = false,
			locked = false,
			shown = true,
		}
		db.pos = db.pos or {
			point = "CENTER",
			xOff = 0,
			yOff = -100
		}
		SpeedyGonzalesDB = db
		self:Initialize()
		self:UnregisterEvent("ADDON_LOADED")
	end
end


function addon:PLAYER_ENTERING_WORLD()
	if UnitInVehicle("player") then
		speeder = "vehicle"
	end
end


function addon:UNIT_ENTERED_VEHICLE(unit)
	if unit == "player" then
		speeder = "vehicle"
	end
end


function addon:UNIT_EXITED_VEHICLE(unit)
	if unit == "player" then
		speeder = "player"
	end
end


function addon:Initialize()
	for i, button in ipairs(optionsFrame.options) do
		button:SetChecked(db[button.setting])
	end
	
	local function onClick(self)
		UIDropDownMenu_SetSelectedValue(dropdown, self.value)
		db.units = self.value
		addon:FixWidth()
	end
	
	local info = {}
	local items = {
		{
			text = "Percent",
			value = "percent",
		},
		{
			text = "Yards per second",
			value = "yards",
		},
		{
			text = "Miles per hour",
			value = "miles",
		},
		{
			text = "Kilometers per hour",
			value = "kilometers",
		},
		{
			text = "Meters per second",
			value = "meters",
		},
	}
	
	UIDropDownMenu_Initialize(dropdown, function(self)
		for i, v in ipairs(items) do
			info.text = v.text
			info.value = v.value
			info.func = onClick
			UIDropDownMenu_AddButton(info)
			info.checked = nil
		end
	end)
	UIDropDownMenu_SetWidth(dropdown, 120)
	UIDropDownMenu_SetSelectedValue(dropdown, db.units)
	
	self:SetPosition()
	self:SetVisibility()
	self:SetLock()
	self:SetPitchDisplay()
	self:FixWidth()
end


function addon:GetSpeed()
	local speed = GetUnitSpeed(speeder)
	
	if db.units == "percent" then
		speed = floor(speed / 7 * 100 + 0.1)
	elseif db.units == "yards" then
		speed = floor(speed * 10 + 0.01) / 10
	elseif db.units == "miles" then
		speed = floor(speed / 1.76 * 36 + 0.01) / 10
	elseif db.units == "kilometers" then
		speed = floor(speed * 9.144 * 3.6 + 0.01) / 10
	elseif db.units == "meters" then
		speed = floor(speed * 9.144 + 0.01) / 10
	end
	
	return speed
end


function addon:OnMouseUp()
	self:StopMovingOrSizing()
	local point, _, _, xOff, yOff = self:GetPoint()
	local pos = db.pos
	pos.point = point
	pos.xOff = xOff
	pos.yOff = yOff
end


function addon:SetPosition()
	local pos = db.pos
	self:SetPoint(pos.point, pos.xOff, pos.yOff)
end


function addon:SetVisibility()
	if db.shown then
		self:Show()
	else
		self:Hide()
	end
end


function addon:SetLock()
	self:EnableMouse(not db.locked)
end


-- resize frame so the pitch display fits
function addon:SetPitchDisplay()
	if db.pitch then
		self:SetHeight(48)
	else
		self:SetHeight(32)
	end
end


-- set width depending on displayed unit type
function addon:FixWidth()
	self:SetWidth(unitData[db.units].width)
end